/* file pqueue.h
 * unstable version. supressed prev, nget and nputs for heads
 * inserted pqueue_close
 * SunOS zape 5.5.1 Generic_103640-12 sun4m sparc SUNW,SPARCstation-10
 * Based in the implementation of a parallel queue by CWK 3/97 
 *
 * output:
 *   void pqueue_put( pqueue_type *pq, char *d )
 *   char * pqueue_get ( pqueue_type *pq )
 *   void pqueue_init( pqueue_type **ppq ) 
 *   void pqueue_close( pqueue_type **ppq )
 *   void pqueue_error(); 
 */

typedef struct pqueue_el {
   struct pqueue_el *next;
   char * data; 
} pqueue_element;
   
typedef struct {
   pqueue_element *first, *last;
   fair_lock lock;
} pqueue_head;


typedef struct {
  int numberofqueues; /* numberofqueues is provided by the user */
  pqueue_head **head;  /*array of pointers to sublist heads*/
  int ngets, nputs;
  int size;
} pqueue_type;

void pqueue_error(void) {
  pprintf("\nFatal error. Not enough shared heap memory\n");
  exit(1);
}

void pqueue_init( pqueue_type **ppq ) {
   pr int i,np;

   barrier;
   np = async_groupsize();
   if (!$) { /*mask out all but one processor*/
     if (*ppq = (pqueue_type*) shmalloc(sizeof( pqueue_type ))) {
      (*ppq)->numberofqueues = np;
      if (!((*ppq)->head = 
					 (pqueue_head**) shmalloc(np*sizeof(pqueue_head *)))) 
						 pqueue_error(); 
     }
     else pqueue_error();
    }
   barrier;
    if (!((*ppq)->head[$] = 
					 (pqueue_head*)shmalloc(sizeof(pqueue_head)))) 
							pqueue_error();
    (*ppq)->head[$]->first = (*ppq)->head[$]->last = NULL;
    fair_lock_init( &((*ppq)->head[$]->lock) );
   (*ppq)->ngets = (*ppq)->nputs = (*ppq)->size = 0;
   barrier;
}

void pqueue_close( pqueue_type **ppq ) {
   pr int i,np;

  barrier;
	shfree((*ppq)->head[$]);
	barrier;
	 if (!$) { /*mask out all but one processor*/
			shfree((*ppq)->head);
			shfree(*ppq);
			*ppq = NULL;
	}
	barrier;
}

void pqueue_put( pqueue_type *pq, char *d )
{
 pr int myputidx;
 pr pqueue_head *mylist;
 pr pqueue_element *e;

  if (e = (pqueue_element *) shmalloc(sizeof(pqueue_element))) {
    e->data = d;
    myputidx = mpadd(&(pq->nputs), 1);
    mylist = pq->head[myputidx % pq->numberofqueues];
    e->next = NULL;
    fair_lockup( &(mylist->lock) );
     if (mylist->last) mylist->last->next = e;
     else mylist->first = e;
     mylist->last = e;
    fair_unlock( &(mylist->lock) );
    /* the Kessler version incremetns size before the item is inserted,
    that is why their get operation has to wait for the puts operations */
    syncadd(&(pq->size), 1);
  }
  else pqueue_error();
}

char * pqueue_get ( pqueue_type *pq, int * extracted )
{
 pr int mygetidx;
 pr pqueue_head *mylist;
 pr pqueue_element *e;
 pr char * d;

 if (tdxr(&(pq->size),extracted,1,0)) { /*test-decrement-exchange-retest;*/
   mygetidx = mpadd( &(pq->ngets), 1);
   mylist = pq->head[mygetidx % pq->numberofqueues];
   fair_lockup( &(mylist->lock) );
     e = mylist->first;
     mylist->first = e->next;
     if (!(e->next)) mylist->last = NULL; 
   fair_unlock( &(mylist->lock) );
	 d = e->data;
   shfree(e);
   return d;
 }
 else return NULL;
}


